#!/usr/bin/python

# Space Bees Attack!
# Entry for Ludum Dare 23 Jam
# Copyright Dylan McCall <http://www.dylanmccall.com>

# Space Bees Attack! is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Space Bees Attack! is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Space Bees Attack!.  If not, see <http://www.gnu.org/licenses/>.

import sys
import math
import random

import time

from pyglet import *
from pyglet.gl import *

import pymunk

import cocos
from cocos.director import director

#TODO: We should really be using batch rendering. We aren't.
#TODO: Switch (completely) to pymunk's Vec2d class instead of coordinate tuples.
#TODO: Transition from Pyglet to Cocos where appropriate. Right now we just have a thin layer of Cocos because I want CocosNode

SCREEN_SIZE = (1200, 720)

'''
Some object in space, with an attached pymunk body, shape or more.
'''
class Entity(object):
	class Events(pyglet.event.EventDispatcher):
		pass
	
	def __init__(self):
		self.is_alive = True
		self.events = Entity.Events()
	
	def get_current_position(self):
		raise NotImplementedError()
	
	def get_current_angle(self):
		raise NotImplementedError()
	
	def update(self):
		pass
	
	def draw(self):
		raise NotImplementedError()
	
	def die(self):
		if self.is_alive:
			self.is_alive = False
			self.events.dispatch_event("on_die")
			self.on_die()
	
	def on_die(self):
		pass

Entity.Events.register_event_type('on_die')

class SpriteEntity(Entity):
	def __init__(self, sprite = None, img = None, img_path = None, anchor = None):
		super(SpriteEntity, self).__init__()
		if sprite:
			self.set_sprite(sprite)
		elif img:
			self.set_image(img, anchor)
		elif img_path:
			self.set_image_path(img_path, anchor)
	
	def set_sprite(self, sprite):
		self.sprite = sprite
		self.size = (self.sprite.width, self.sprite.height)
		self.center = (self.size[0]/2, self.size[1]/2)
		self.radius = max(self.sprite.width, self.sprite.height)
	
	def set_image(self, img, anchor = None):
		if anchor:
			img.anchor_x = anchor[0]
			img.anchor_y = anchor[1]
		else:
			img.anchor_x = img.width/2
			img.anchor_y = img.height/2
		img_sprite = sprite.Sprite(img)
		self.set_sprite(img_sprite)
	
	def set_image_path(self, img_path, anchor = None):
		img = image.load(img_path)
		self.set_image(img, anchor)
	
	def draw(self):
		self.update()
		
		position = self.get_current_position()
		angle = self.get_current_angle()
		
		self.sprite.x = position[0]
		self.sprite.y = position[1]
		self.sprite.rotation = angle
		
		self.sprite.draw()

class PhysicsEntity(SpriteEntity):
	def __init__(self, space, position = (0,0), mass = None, inertia = None, **kwargs):
		super(PhysicsEntity, self).__init__(**kwargs)
		
		if mass and inertia:
			self.pm_body = pymunk.Body(mass, inertia)
		else:
			self.pm_body = pymunk.Body()
		self.pm_body.position = (position[0]+self.center[0], position[1]+self.center[1])
		
		self.space = space
		
		clock.schedule_interval(self._test_off_screen, 1)
	
	def _test_off_screen(self, dt):
		position = self.get_current_position()
		off_screen = position[0] + self.radius < 0
		off_screen = off_screen or position[0] - self.radius > self.space.size[0]
		off_screen = off_screen or position[1] + self.radius < 0
		off_screen = off_screen or position[1] - self.radius > self.space.size[1]
		if off_screen:
			self.die()
	
	def get_current_position(self):
		if self.pm_body:
			return (self.pm_body.position.x, self.pm_body.position.y)
		else:
			return (0,0)
	
	def get_current_angle(self):
		if self.pm_body:
			return math.degrees(-self.pm_body.angle)
		else:
			return 0

class GroundEntity(PhysicsEntity):
	def __init__(self, space):
		position = (50, 0)
		size = (space.size[0]-100, 140)
		
		ground_pattern = image.SolidColorImagePattern((100, 200, 20, 255))
		ground_image = ground_pattern.create_image(size[0], size[1])
		
		super(GroundEntity, self).__init__(space, position=position, img=ground_image)
		
		self.pm_shape = pymunk.Poly.create_box(self.pm_body, size)
		self.pm_shape.elasticity = 0.0
		self.pm_shape.friction = 0.99
		self.space.add_static(self.pm_shape)
	
	def _test_off_screen(self, dt):
		pass

# Very important ... thing.
class VITEntity(PhysicsEntity):
	COLLISION_TYPE = 100
	
	FORM_LIVING = 1
	FORM_DAMAGED = 2
	FORM_BROKEN = 3
	
	def __init__(self, space, position):
		self.form = VITEntity.FORM_LIVING
		
		self.initial_damage = 800
		self.damage = self.initial_damage
		
		base_size = 64
		mass = 2
		inertia = pymunk.moment_for_box(mass, base_size, base_size)
		
		super(VITEntity, self).__init__(space, mass=mass, inertia=inertia, position=position, img_path="graphics/happy-thing.png")
		
		self.pm_shape = pymunk.Poly.create_box(self.pm_body, (base_size, base_size))
		self.pm_shape.elasticity = 0.5
		self.pm_shape.friction = 0.5
		#self.pm_shape.collision_type = VITEntity.COLLISION_TYPE
		
		self.pm_shape.sb_event_handler = self
		space.add(self.pm_body, self.pm_shape)
	
	def collision_after_cb(self, space, arbiter):
		if arbiter.total_impulse.length > 25 and arbiter.is_first_contact:
			self.damage -= arbiter.total_impulse.length
	
	def update(self):
		super(VITEntity, self).update()
		
		if not self.pm_body:
			return
		
		current_position = int(self.pm_body.position.x), int(self.pm_body.position.y)
		self.sprite.rotation = math.degrees(-self.pm_body.angle)
		self.sprite.x = current_position[0]
		self.sprite.y = current_position[1]
		
		if self.damage <= 0:
			self.die()
		
		if self.damage < (self.initial_damage*0.75) and self.form < VITEntity.FORM_DAMAGED:
			self.set_image_path("graphics/damaged-happy-thing.png")
			self.form = VITEntity.FORM_DAMAGED
		
		if not self.is_alive and self.form < VITEntity.FORM_BROKEN:
			self.set_image_path("graphics/broken-happy-thing.png", anchor=(32,22))
			
			# New body will start with its own momentum, conveniently giving a "dropping like a stone" effect
			# I'll just say I meant it like this :b
			mass = 1
			inertia = pymunk.moment_for_box(mass, 64, 43)
			new_body = pymunk.Body(mass, inertia)
			new_body.position = self.pm_body.position
			new_body.angle = self.pm_body.angle
			
			new_shape = pymunk.Poly.create_box(new_body, (64, 43))
			new_shape.elasticity = 0.01
			new_shape.friction = 0.99
			
			old_pm_body = self.pm_body
			old_pm_shape = self.pm_shape
			self.pm_body = None
			self.pm_shape = None
			self.space.remove(old_pm_body, old_pm_shape)
			
			self.pm_body = new_body
			self.pm_shape = new_shape
			self.space.add(self.pm_body, self.pm_shape)
			
			self.form = VITEntity.FORM_BROKEN
			
			#TODO: Throw in a particle effect!

class AttackerEntity(PhysicsEntity):
	def __init__(self, space, position, targets = []):
		mass = 0.01
		inertia = pymunk.moment_for_box(mass, 33, 23)
		
		super(AttackerEntity, self).__init__(space, mass=mass, inertia=inertia, position=position, img_path="graphics/spacebee.png")
		
		vertices = [
			(0,0),
			(16,0),
			(33,12),
			(16,23),
			(0,23)
		]
		
		self.pm_shape = pymunk.Poly(self.pm_body, vertices, offset=(-33/2, -23/2))
		self.pm_shape.elasticity = 0.05
		self.pm_shape.friction = 0.7
		self.pm_shape.group = SBSpace.COLGROUP_ATTACKERS
		
		self.pm_shape.sb_event_handler = self
		self.space.add(self.pm_body, self.pm_shape)
		
		self.move_body = pymunk.Body()
		self.turn_constraint = pymunk.RotaryLimitJoint(self.move_body, self.pm_body, 0, 0)
		self.turn_constraint.error_bias = 0#pow(1.0 - 0.1, 30.0)
		self.space.add(self.turn_constraint)
		
		self.targets = targets
		self.set_random_target()
		self.do_attack = False
		self.attack_strength = 0
		
		clock.schedule_interval(self.set_random_target_interval, 5)
		clock.schedule_interval(self.random_attack_interval, 12)
	
	def _test_off_screen(self, dt):
		pass
	
	def collision_before_cb(self, space, arbiter):
		if self.is_alive:
			return True
		else:
			return False
	
	def collision_after_cb(self, space, arbiter):
		if arbiter.total_impulse.length > 5 and len(arbiter.shapes) > 0:
			self.do_attack = True
			self.attack_strength = 4000
	
	def on_die(self):
		super(AttackerEntity, self).on_die()
		self.pm_body.mass = 8
		self.pm_body.moment = pymunk.moment_for_box(self.pm_body.mass, 33, 23)
	
	def set_random_target(self):
		if random.randint(0,5) == 0 and len(self.targets) > 0:
			self.target_point = random.choice(self.targets)
		else:
			target_x = random.randint(0, self.space.size[0])
			target_y = random.randint(150, self.space.size[1]) #140 is the height of the ground
			self.target_point = (target_x, target_y)
	
	def set_random_target_interval(self, dt):
		self.set_random_target()
	
	def random_attack_interval(self, dt):
		if random.randint(0,15) == 0:
			self.do_attack = True
			self.attack_strength = -2500
	
	def attack(self):
		self.space.explode_from_point((self.pm_body.position.x, self.pm_body.position.y), self.attack_strength, [self.pm_body])
		self.do_attack = False
		self.die()
	
	def update(self):
		super(AttackerEntity, self).update()
		
		# Apply force and rotate towards target point!
		if self.is_alive and self.do_attack:
			self.attack()
		
		if self.is_alive:
			#FIXME: Turn towards target, apply force in current facing direction!
			offset_y = self.target_point[1] - self.pm_body.position.y
			offset_x = self.target_point[0] - self.pm_body.position.x
			angle_to_target = math.atan2(offset_y, offset_x)
			
			self.turn_constraint.min = self.turn_constraint.max = angle_to_target
			
			angle_offset = abs(angle_to_target - self.pm_body.angle)
			distance = math.hypot(offset_y, offset_x)
			if angle_offset < math.pi/24 and distance < 400:
				move_power = 0.1
			else:
				move_power = 0.025
			force_x = move_power * math.cos(self.pm_body.angle)
			force_y = move_power * math.sin(self.pm_body.angle)
			force_y += 0.05
			self.pm_body.apply_impulse((force_x, force_y), (0,0))



class AttackWaveMember(object):
	def __init__(self, angle, distance, targets = []):
		self.angle = angle
		self.distance = distance
		self.targets = targets
	
	def realize(self, space):
		shift_x = math.cos(self.angle) * self.distance
		shift_y = (math.sin(self.angle) * self.distance)+300
		
		spawn_point = (shift_x, shift_y)
		
		attacker = AttackerEntity(space, spawn_point, self.targets)
		attacker.pm_body.angle = self.angle
		return attacker

class AttackWave(object):
	def __init__(self):
		self.members = []
	
	def add_member(self, member):
		self.members.append(member)
	
	def realize(self, space):
		attackers = []
		for member in self.members:
			new_attacker = member.realize(space)
			attackers.append(new_attacker)
		return attackers

class Cell(object):
	class Sprite(cocos.sprite.Sprite):
		def __init__(self, cell, block):
			self.cell = cell
			self.block = block
			
			type_data = Cell.TYPE_DATA[cell.cell_type]
			
			img = image.load(type_data['img_name'])
			img.anchor_x = img.width/2
			img.anchor_y = img.height/2
			
			super(Cell.Sprite, self).__init__(img)
			
			self.sprite = cocos.sprite.Sprite(img)
		
		def draw(self):
			#Is parent "real"? Change opacity!
			self.opacity = 255 if self.block.is_real else 200
			super(Cell.Sprite, self).draw()
	
	BASE_SIZE = 48
	
	TYPE_NULL = 0
	TYPE_NORMAL = 1
	TYPE_GLASS = 2
	TYPE_FIRE = 3
	TYPE_STICKY = 4
	TYPE_STRONG = 5
	
	TYPE_DATA = {
		TYPE_NORMAL : {
			'img_name' : 'graphics/block-grey.png',
			'damage' : 3500,
			'elasticity' : 0,
			'friction' : 0.99,
			'on_collide' : None
		},
		TYPE_GLASS : {
			'img_name' : 'graphics/block-blue.png',
			'damage' : 2000,
			'elasticity' : 0,
			'friction' : 0.8,
			'on_collide' : None
		},
		TYPE_FIRE : {
			'img_name' : 'graphics/block-red.png',
			'damage' : 3500,
			'elasticity' : 0,
			'friction' : 0.99,
			'on_collide' : None
		},
		TYPE_STICKY : {
			'img_name' : 'graphics/block-purple.png',
			'damage' : 3500,
			'elasticity' : 0,
			'friction' : 0.99,
			'on_collide' : None
		},
		TYPE_STRONG : {
			'img_name' : 'graphics/block-black.png',
			'damage' : 6000,
			'elasticity' : 0,
			'friction' : 0.99,
			'on_collide' : None
		}
	}
	
	def __init__(self, block, position, offset = (0,0), cell_type = TYPE_NORMAL):
		size = Cell.BASE_SIZE
		self.block = block
		self.cell_type = cell_type
		
		type_data = Cell.TYPE_DATA[cell_type]
		
		self.initial_damage = type_data['damage']
		self.damage = self.initial_damage
		
		# Get distance from center
		position_x = (position[0] * -size) - offset[0]
		position_y = (position[1] * -size) - offset[1]
		
		self.sprite = Cell.Sprite(self, block)
		self.sprite.position = (position_x, position_y)
		self.block.root_node.add(self.sprite)
		
		vertices = [
			(0,0),
			(size,0),
			(size,size),
			(0,size)
		]
		
		self.pm_shape = pymunk.Poly(block.pm_body, vertices, offset=(position_x-(size/2), position_y-(size/2)))
		self.pm_shape.elasticity = type_data['elasticity']
		self.pm_shape.friction = type_data['friction']
		self.pm_shape.sb_event_handler = self
		
		if not block.is_real:
			self.pm_shape.group = SBSpace.COLGROUP_ATTACKERS
			self.pm_shape.collision_type = Block.FAKE_COLLISION_TYPE
	
	def make_real(self):
		self.pm_shape.group = 0
		self.pm_shape.collision_type = 0
	
	def collision_before_cb(self, space, arbiter):
		if self.cell_type == Cell.TYPE_NULL:
			return False
		else:
			return True
	
	def collision_after_cb(self, space, arbiter):
		# detect collision
		# emit damage according to cell type
		# kill self according to cell type's strength
		# FIXME: block needs to recalculate mass
		if arbiter.total_impulse.length > 25 and arbiter.is_first_contact:
			self.damage -= arbiter.total_impulse.length
	
	def update(self):
		if self.damage <= 0 and self.cell_type > Cell.TYPE_NULL:
			self.block.root_node.remove(self.sprite)
			self.sprite = None
			self.cell_type = Cell.TYPE_NULL
		#else:
		#	if self.damage < self.initial_damage:
		#		self.damage += min(100, self.initial_damage-self.damage)

'''
A Tetris-style block, with arbitrary rows and columns
'''
class Block(PhysicsEntity):
	FAKE_COLLISION_TYPE = 100
	
	def __init__(self, space, position = (0, 0), pattern = [[1]], is_real = True):
		self.cells = []
		self.rotation = 0 # Rotation in 90 degree (pi/2) intervals
		self.is_real = is_real
		
		total_width = 0
		total_height = 0
		for y,row in enumerate(pattern):
			for x,cell_type in enumerate(row):
				if cell_type > 0:
					total_width = max(total_width, x)
					total_height = max(total_height, y)
		self.size = (total_width * Cell.BASE_SIZE, total_height * Cell.BASE_SIZE)
		self.center = (self.size[0] / 2, self.size[1] / 2)
		self.radius = max(self.size[0], self.size[1])
		
		mass = 3 if is_real else 0.006
		inertia = self._calculate_inertia(mass)
		
		super(Block, self).__init__(space, mass=mass, inertia=inertia, position=position)
		
		self.root_node = cocos.cocosnode.CocosNode()
		
		for y,row in enumerate(pattern):
			for x,cell_type in enumerate(row):
				if cell_type > 0:
					cell = Cell(self, (x, y), (-self.center[0], -self.center[1]), cell_type)
					self.cells.append(cell)
		
		shapes = [cell.pm_shape for cell in self.cells]
		self.space.add(self.pm_body, shapes)
	
	def set_sprite(self):
		#This isn't REALLY a SpriteEntity. Just a crappy class hierarchy.
		raise NotImplementedError()
	
	def make_real(self):
		self.is_real = True
		
		mass = 3
		self.pm_body.mass = mass
		self.pm_body.moment = self._calculate_inertia(mass)
		
		for cell in self.cells:
			cell.make_real()
	
	def _calculate_inertia(self, mass):
		return pymunk.moment_for_box(mass, self.size[0], self.size[1])
		
	def on_die(self):
		super(Block, self).on_die()
		
		old_pm_body = self.pm_body
		self.pm_body = None
		
		shapes = []
		for cell in self.cells:
			shapes.append(cell.pm_shape)
			cell.pm_shape = None
		
		self.space.remove(old_pm_body, shapes)
	
	def overlapping_shapes(self):
		results = []
		for cell in self.cells:
			cell_results = self.space.shape_query(cell.pm_shape)
			results.extend(cell_results)
		return results
	
	def draw(self):
		if self.pm_body:
			self.root_node.rotation = math.degrees(-self.pm_body.angle)
			self.root_node.position = int(self.pm_body.position.x), int(self.pm_body.position.y)
			
			self.root_node.visit()
			
			for cell in self.cells:
				cell.update()

class BlockGenerator(object):
	BLOCK_PATTERNS = [
		# Classic Tetris blocks
		[[1,1,1,1]],
		[[1],[1,1,1]],
		[[0,0,1],[1,1,1]],
		[[1,1],[1,1]],
		[[0,1,1],[1,1,0]],
		[[0,1,0],[1,1,1]],
		[[1,1,0],[0,1,1]],
		# Other blocks
		[[1,0,1],[1,1,1]],
	]
	
	@staticmethod
	def get_random_pattern():
		pattern = random.choice(BlockGenerator.BLOCK_PATTERNS)
		
		type_choices = [
			(Cell.TYPE_NORMAL, 6/10.0),
			(Cell.TYPE_GLASS, 0.5/10.0),
			(Cell.TYPE_FIRE, 1/10.0),
			(Cell.TYPE_STICKY, 1.5/10.0),
			(Cell.TYPE_STRONG, 1/10.0)
		]
		
		for y,row in enumerate(pattern):
			for x,cell_type in enumerate(row):
				if cell_type > 0:
					random_type = BlockGenerator._w_choice(type_choices)
					pattern[y][x] = random_type
		
		return pattern
	
	#Via <http://snippets.dzone.com/posts/show/732>
	@staticmethod
	def _w_choice(lst):
		n = random.uniform(0, 1)
		for item, weight in lst:
			if n < weight:
				break
			n = n - weight
		return item

class BlockPlacer(object):
	class Events(pyglet.event.EventDispatcher):
		pass
	
	def __init__(self, space, on_release_cb):
		self.events = BlockPlacer.Events()
		self.events.set_handler("on_release", on_release_cb)
		
		self.space = space
		
		self.do_release = False
		
		pattern = BlockGenerator.get_random_pattern()
		self.block = Block(space, (500,800), pattern, is_real = False)
		
		self.move_body = pymunk.Body()
		
		self.key_state = window.key.KeyStateHandler()
		
		self.move_constraint = pymunk.PivotJoint(self.move_body, self.block.pm_body, self.block.pm_body.position, (0, 0))
		self.move_constraint.error_bias = 0
		
		self.rotate_constraint = pymunk.RotaryLimitJoint(self.move_body, self.block.pm_body, 0, 0)
		self.rotate_constraint.error_bias = pow(1.0 - 0.1, 50.0)
		
		space.add(self.move_constraint, self.rotate_constraint)
		
		space.add_collision_handler(Block.FAKE_COLLISION_TYPE, 0, pre_solve = self.on_block_placer_collision_pre_solve, post_solve = self.on_block_placer_collision_post_solve)
		
		self.block.events.set_handler("on_die", self.on_block_died)
	
	def on_block_placer_collision_pre_solve(self, space, arbiter):
		return True
	
	def on_block_placer_collision_post_solve(self, space, arbiter):
		#TODO: Swap these when attackers are on a different pymunk layer
		if arbiter.total_impulse.y > 25:
			self.do_release = True
		#for contact in arbiter.contacts:
		#	if contact.distance * contact.normal.y > 2:
		#		self.do_release = True
	
	def push_point(self, position):
		if not self.finished:
			self.points.append(position)
		self.finished = len(self.points) >= 2
	
	def is_finished(self):
		return self.finished
	
	def try_move_x(self, value):
		# try to move on the X axis, but don't do it if it's impossible
		target_x = self.move_constraint.anchr1.x + value
		min_x = 0-self.block.center[0]
		max_x = self.space.size[0]+self.block.center[0]
		#if target_x > min_x and target_x < max_x:
		self.move_constraint.anchr1.x = target_x
	
	def try_rotate(self, spin):
		# get current spin (angle snapped to pi/2)
		current_spin = int(self.rotate_constraint.min / (math.pi/2))
		new_angle = (current_spin + spin) * (math.pi/2)
		self.rotate_constraint.min = self.rotate_constraint.max = new_angle % (math.pi*5)
	
	def on_block_died(self):
		self.release_block()
	
	def release_block(self):
		if self.move_constraint and self.rotate_constraint:
			old_move_constraint = self.move_constraint
			old_rotate_constraint = self.rotate_constraint
			self.move_constraint = None
			self.rotate_constraint = None
			self.space.remove(old_move_constraint, old_rotate_constraint)
			self.block.make_real()
			self.events.dispatch_event("on_release", self.block)
	
	def on_key_press(self, key, modifiers):
		self.key_state.on_key_press(key, modifiers)
		
		if key == window.key.SPACE:
			self.do_release = True
		elif key == window.key.UP:
			self.try_rotate(1)
	
	def on_key_release(self, key, modifiers):
		self.key_state.on_key_release(key, modifiers)
	
	def draw(self):
		self.block.draw()
		
		if self.do_release:
			self.release_block()
		else:
			fall_speed = 1
			
			if self.key_state[window.key.LEFT]:
				self.try_move_x(-5)
			elif self.key_state[window.key.RIGHT]:
				self.try_move_x(+5)
			elif self.key_state[window.key.DOWN]:
				fall_speed = 2
			
			self.move_constraint.anchr1.y -= fall_speed
			
			if self.move_constraint.anchr1.y < 0:
				self.do_release = True

BlockPlacer.Events.register_event_type("on_release")



class GameHUD():
	def __init__(self, size):
		self.size = size
		
		self.fps_display = clock.ClockDisplay()
		
		self.lives_label = pyglet.text.Label(x=10, y=size[1]-28)
		self.lives_label_state = 0
		
		overlay_pattern = image.SolidColorImagePattern((0, 0, 0, 180))
		overlay_image = overlay_pattern.create_image(size[0], size[1])
		self.overlay_background = sprite.Sprite(overlay_image)
		self.overlay_label = pyglet.text.Label(font_size=36, x=0, y=size[1]/2, width=size[0], multiline=True)
		self.overlay_label.set_style("align","center")
		
		self.overlay_visible = False
	
	def start_game(self):
		self.lives_label.text = "Defend the happy boxes!"
		self.lives_label_state = 0
		clock.unschedule(self.blink_lives_label_interval)
		
		self.overlay_visible = False
	
	def update_good_guys(self, count, blink = True):
		if count == 1:
			self.lives_label.text = "%d happy box" % count
		else:
			self.lives_label.text = "%d happy boxes" % count
		
		if blink:
			clock.schedule_interval(self.blink_lives_label_interval, 0.5)
			clock.schedule_once(self.stop_blink_lives_label_interval, 5)
	
	def stop_blink_lives_label_interval(self, dt):
		self.lives_label_state = 0
		clock.unschedule(self.blink_lives_label_interval)
	
	def blink_lives_label_interval(self, dt):
		self.lives_label_state = (self.lives_label_state + 1) % 2
	
	def show_initial_help(self):
		self.overlay_label.text = u"Use the arrow keys to guide the blocks that fall from the top of the screen.\nBuild structures to protect your innocent, happy and fragile boxes from the evil space bees!\n\nPress Return \u23CE to start the game."
		self.overlay_label.font_size=14
		self.overlay_visible = True
	
	def show_gameover(self, survival_time):
		self.overlay_label.text = u"Game over :(\nYou survived %d seconds.\nPress Return \u23CE to start again" % int(survival_time)
		self.overlay_label.font_size=18
		self.overlay_visible = True
	
	def draw(self):
		if self.lives_label_state == 0:
			self.lives_label.color = (255, 255, 255, 255)
		else:
			self.lives_label.color = (255, 116, 77, 255)
		
		self.lives_label.draw()
		#self.fps_display.draw()
		
		if self.overlay_visible:
			self.overlay_background.draw()
			self.overlay_label.draw()

class SBSpace(pymunk.Space):
	COLGROUP_ATTACKERS = 100
	
	def __init__(self, size, *args, **kwargs):
		super(SBSpace, self).__init__(*args, **kwargs)
		
		self.gravity = (0.0, -250)
		self.set_default_collision_handler(pre_solve = self.on_any_collision_pre_solve, post_solve = self.on_any_collision_post_solve)
		self.sleep_time_threshold = 5
		
		self.size = size
	
	def on_any_collision_pre_solve(self, space, arbiter):
		value = True
		for shape in arbiter.shapes:
			if hasattr(shape, "sb_event_handler") and hasattr(shape.sb_event_handler, "collision_before_cb"):
				cb_return = shape.sb_event_handler.collision_before_cb(space, arbiter)
				if cb_return != None:
					value = value and cb_return
		return value
	
	def on_any_collision_post_solve(self, space, arbiter):
		value = True
		for shape in arbiter.shapes:
			if hasattr(shape, "sb_event_handler") and hasattr(shape.sb_event_handler, "collision_after_cb"):
				cb_return = shape.sb_event_handler.collision_after_cb(space, arbiter)
				if cb_return != None:
					value = value and cb_return
		return value
	
	#TODO: Every 1s, cull anything that falls off the screen.
	
	def explode_from_point(self, center, intensity, ignore = []):
		for body in self.bodies:
			if body not in ignore:
				offset_x = body.position.x - center[0]
				offset_y = body.position.y - center[1]
				angle = math.atan2(offset_y, offset_x)
				distance = math.hypot(offset_x, offset_y)
			
				if distance > 0:
					force = intensity / distance #FIXME: this should have a logarithmic falloff
					force_x = -force * math.cos(angle)
					force_y = -force * math.sin(angle)
					body.apply_impulse((force_x, force_y), (0,0))

class SpaceBees():
	GAME_STATE_INITIAL = 0
	GAME_STATE_RUNNING = 1
	GAME_STATE_STOPPED = 2
	
	def __init__(self, size):
		self.size = size
		
		self.state = SpaceBees.GAME_STATE_INITIAL
		self.hud = GameHUD(size)
		
		self.hud.show_initial_help()
		
		self.menu_key_pressed = False
	
	def start_game(self):
		self.state = SpaceBees.GAME_STATE_RUNNING
		
		self.start_time = time.time()
		
		self.hud.start_game()
		
		self.space = SBSpace(self.size)
		
		sky_pattern = image.SolidColorImagePattern((132, 174, 186, 255))
		sky_image = sky_pattern.create_image(self.size[0], self.size[1])
		self.sky_sprite = sprite.Sprite(sky_image)
		
		self.ground = GroundEntity(self.space)
		self.entities = []
		
		self.good_guys = []
		self.attack_waves = []
		self.bad_guys = []
		
		self.block_placer = None
		clock.schedule_once(self.add_block, 2)
		
		self.add_good_guy(VITEntity(self.space, (320, 148)))
		self.add_good_guy(VITEntity(self.space, (800, 148)))
		
		self.attack_wave_number = 1
		clock.schedule_interval(self.attack_wave_interval, 20)
	
	def do_gameover(self):
		if self.state == SpaceBees.GAME_STATE_RUNNING:
			survival_time = time.time() - self.start_time
			
			self.hud.show_gameover(survival_time)
			self.state = SpaceBees.GAME_STATE_STOPPED
			
			clock.unschedule(self.add_block)
			clock.unschedule(self.attack_wave_interval)
	
	def add_random_attack_wave(self, count):
		wave = AttackWave()
		
		for n in range(count):
			angle = random.uniform(0.05,math.pi-0.05)
			distance = random.randint(1000, 1500)
			targets = []
			for entity in self.good_guys:
				if entity.is_alive:
					targets.append(entity.get_current_position())
			attacker = AttackWaveMember(angle, distance, targets)
			wave.add_member(attacker)
		
		new_bad_guys = wave.realize(self.space)
		self.bad_guys.extend(new_bad_guys)
	
	def attack_wave_interval(self, dt):
		#Remove existing attackers
		for entity in self.bad_guys:
			entity.die()
		self.add_random_attack_wave(self.attack_wave_number*2)
		self.attack_wave_number += 1
	
	def add_good_guy(self, entity):
		if self.state == SpaceBees.GAME_STATE_RUNNING:
			self.good_guys.append(entity)
			entity.events.set_handler("on_die", self.on_good_guy_died)
	
	def on_good_guy_died(self):
		if self.state == SpaceBees.GAME_STATE_RUNNING:
			living_good_guys = 0
			for entity in self.good_guys:
				if entity.is_alive:
					living_good_guys += 1
			self.hud.update_good_guys(living_good_guys)
			if living_good_guys == 0:
				self.do_gameover()
	
	def add_block(self, dt):
		if self.state == SpaceBees.GAME_STATE_RUNNING:
			self.block_placer = BlockPlacer(self.space, self.on_block_placer_released)
	
	def on_block_placer_released(self, new_block):
		if self.state == SpaceBees.GAME_STATE_RUNNING:
			self.entities.append(new_block)
			self.block_placer = None
			clock.schedule_once(self.add_block, 2)
	
	def on_key_press(self, key, modifiers):
		if self.state == SpaceBees.GAME_STATE_RUNNING:
			if self.block_placer:
				self.block_placer.on_key_press(key, modifiers)
		elif key == window.key.RETURN:
			self.menu_key_pressed = True
	
	def on_key_release(self, key, modifiers):
		if self.state == SpaceBees.GAME_STATE_RUNNING:
			if self.block_placer:
				self.block_placer.on_key_release(key, modifiers)
		elif key == window.key.RETURN:
			self.menu_key_pressed = False
			self.start_game()
	
	def draw(self, step):
		if self.state > SpaceBees.GAME_STATE_INITIAL:
			self.space.step(step)
			self.sky_sprite.draw()
			self.ground.draw()
			
			for entity_list in self.entities, self.good_guys, self.bad_guys:
				for entity in entity_list:
					entity.draw()
			
			if self.state == SpaceBees.GAME_STATE_RUNNING and self.block_placer:
				self.block_placer.draw()
		
		self.hud.draw()

class SpaceBeesRoot(cocos.layer.Layer):
	is_event_handler = True
	
	def __init__(self):
		super(SpaceBeesRoot, self).__init__()
		
		clock.set_fps_limit(60)
		
		self.game = SpaceBees(SCREEN_SIZE)
	
	def on_resize(self, width, height):
		glViewport(0, 0, width, height)
		glMatrixMode(GL_PROJECTION)
		glLoadIdentity()
		glOrtho(0, SCREEN_SIZE[0], 0, SCREEN_SIZE[1], -1, 1)
		glMatrixMode(GL_MODELVIEW)
	
	def draw(self):
		glClear(GL_COLOR_BUFFER_BIT)
		
		self.game.draw(1/60.0)
	
	def on_key_press(self, key, modifiers):
		self.game.on_key_press(key, modifiers)
	
	def on_key_release(self, key, modifiers):
		self.game.on_key_release(key, modifiers)


if __name__ == "__main__":
	director.init(SCREEN_SIZE[0], SCREEN_SIZE[1], resizable=False, caption="Space Bees Attack!")
	
	scene = cocos.scene.Scene()
	scene.add( SpaceBeesRoot() )
	director.run( scene )

